/*
 * eximstate: Exim Information
 * ---------------------------
 *
 * Functions for ascertaining information about Exim's
 * queue.
 *
 * (c) 2002 Claranet UK Ltd. by Ollie Cook <ollie@olliecook.net>
 *
 * $Id: exim_info.c,v 1.4 2002/08/11 15:49:37 ollie Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include "eximstate.h"
#include "exim_info.h"
#include "logging.h"
#include "config.h"

/* Function:    get_queue_count
 * Description: Get the number of messages, and the number of frozen messages, on the queue
 */

int get_queue_count (char *input_dir, int *queue_count, int *frozen_count, int *bounce_count) {
	DIR *directory;
	struct dirent *dirent;
	FILE *filehandle;
	char buf[MAX_BUF];
	char filename[MAX_BUF];
	struct directory_list *dirlist, *temp;
	int rc;

	/* To get started, we need a directory handle */
	debug ("Beginning to build list of spool directories starting at %s", input_dir);
	if ((directory = opendir (input_dir)) == NULL) {
		error ("Failed to opendir(%s)", input_dir);
		return (0);
	}

	/* Create the first element of the linked list of directories we
	 * will be looking at later. This is in case Exim is running with
	 * a split spool directory. If a split spool is not set we will
	 * rather uninterestingly be left with a one element linked list.
	 */
		dirlist = (struct directory_list *)malloc (sizeof (struct directory_list));
		if (dirlist == NULL) {
				error ("Memory allocation error in get_queue_count()");
				return (0);
		}
		debug ("Creating initial entry for %s in spool directory list", input_dir);
		strncpy (dirlist->dir_name, input_dir, sizeof (dirlist->dir_name) - 1);
		dirlist->dir_name[sizeof (dirlist->dir_name)] = '\0';
		dirlist->next = NULL;

	/* Now create the rest of the linked list. */
	debug ("Checking for further directories under %s (split spool)", input_dir);
	rc = create_directory_list (directory, input_dir, &dirlist);
	if (rc == 0) {
		error ("Failed to create linked list of directories under the input directory.");
		return (0);
	}

	closedir (directory);

	/* Now we have a list of directories under the input directory. Go
	 * through each one in turn looking for header files. These are
	 * denoted by having files ending -H.
	 */ 
	temp = dirlist;
	while (dirlist != NULL) {
		debug ("Checking for header files in %s", dirlist->dir_name);

		if ((directory = opendir (dirlist->dir_name)) == NULL) {
			error ("Failed to opendir(%s)", dirlist->dir_name);
			return (0);
		}
	
		while ((dirent = readdir (directory)) != NULL) {
			int linecount = 0;  /* Used to count which line we're on in the header file */

			/* We're only concerned with header files */
			if (dirent->d_name[(strlen (dirent->d_name) - 1)] == 'H'
				&& (dirent->d_type == DT_REG || dirent->d_type == DT_UNKNOWN)) {

				debug ("Found %s in %s", dirent->d_name, dirlist->dir_name);

				/* Open each header file in turn and check it for a line
			 	* beginning '-frozen' which denotes a frozen message
				 */
				snprintf (filename, MAX_BUF, "%s/%s", dirlist->dir_name, dirent->d_name);
				if ((filehandle = fopen (filename, "r")) == NULL) {
					log ("Unable to open message '%s' (%s). Continuing scan", dirent->d_name, strerror (errno)); 
          continue;
				}

				while (fgets (buf, sizeof (buf), filehandle)) {
					linecount++;
	
					/* End of headers? */
					if (!strncmp (buf, "\n", 1))
						break;

					/* Frozen? */
					if (!strncmp (buf, "-frozen", 7)) {
						(*frozen_count)++;
						debug ("  ... message is frozen");
					}
	
					/* Bounce Message? */
					if (!strncmp (buf, "<>", 2)) {
						(*bounce_count)++;
						debug ("  ... message is a bounce");
					}
				}

				fclose (filehandle);
				(*queue_count)++; 
			}
		} 

	dirlist = dirlist->next;
	closedir (directory);

	}

	destroy_directory_list (temp);
	return (1);
}

/* Function:    create_directory_list
 * Description:	Creates a linked list of single letter directories
 *              under the input directory passed.
 */

int create_directory_list (DIR *directory, char *input_dir, struct directory_list **dirlist) {
	struct dirent *dirent;
	struct directory_list *temp;

	/* Iterate through all the directories under the directory we've
	 * been passed, and add the path to the linked list if its length
	 * is 1 character.
	 */

	while ((dirent = readdir (directory)) != NULL) {

		/* We're not interested in '.' and '..' */
		if (*dirent->d_name == '.')
			continue;

		if (dirent->d_type == DT_DIR && strlen (dirent->d_name) == 1) {
			temp = (struct directory_list *)malloc (sizeof (struct directory_list));
			if (temp == NULL) {
				error ("Memory allocation error in create_directory_list()");
				return (0);
			}
			temp->next = *dirlist;
			debug ("Adding %s/%s to spool directory list", input_dir, dirent->d_name);
			snprintf (temp->dir_name, sizeof (temp->dir_name), "%s/%s", input_dir, dirent->d_name);
			*dirlist = temp;
		}
	}

	return (1);
}

/* Function:    destroy_directory_list
 * Description:	Frees up memory associated with a directory_list
 *              linked list.
 */

void destroy_directory_list (struct directory_list *dirlist) {
	struct directory_list *temp;

	while (dirlist != NULL) {
		temp = dirlist;
		dirlist = temp->next;
		free (temp);
	}
	dirlist = NULL;
}
